using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;

using System.Security;
using System.Net;
using System.IO;
using System.Web;
using System.Collections.Specialized;
using System.Text.RegularExpressions;

namespace LJBot
{
	public class LJServer
	{
		/// <summary>
		///        
		/// </summary>
		/// <param name="response"></param>
		/// <returns></returns>
		private StringDictionary MakeDict (string response)
		{
			string[] split = response.Split ("\n".ToCharArray ());

			StringDictionary dict = new StringDictionary ();
			for (int i = 0; i < split.Length - 1; i += 2)
			{
				dict[split[i]] = split[i + 1];
			}

			return dict;
		}

		readonly string ServerApiUri = "http://livejournal.com/interface/flat"; 

		string GetChallenge ()
		{
			string request = "mode=getchallenge";
			string challengeResponse = SendRequest (ServerApiUri, request, null);

			StringDictionary dict = MakeDict (challengeResponse);
			if (dict["success"] == "OK")
			{
				return dict["challenge"];
			}

			return "";
		}

		ILog _log;

		public LJServer (ILog log)
		{
			_log = log;
		}


		/// <summary>
		///   
		/// </summary>
		/// <param name="url"></param>
		/// <returns></returns>
		private string GetPage (string url, CookieCollection cookies)
		{
			HttpWebRequest request =
				(HttpWebRequest)WebRequest.Create (url);

			//   
			//     
			request.AllowAutoRedirect = true;

			request.Credentials = CredentialCache.DefaultCredentials;
			request.Method = "GET";
			request.CookieContainer = new CookieContainer ();

			if (cookies != null)
			{
				request.CookieContainer.Add (cookies);
			}

			//   
			HttpWebResponse response = (HttpWebResponse)request.GetResponse ();

			//  
			Stream responseStream = response.GetResponseStream ();
			StreamReader readStream = new StreamReader (responseStream, Encoding.UTF8);

			string currResponse = readStream.ReadToEnd ();

			readStream.Close ();
			response.Close ();

			return currResponse;
		}

		/// <summary>
		///   cookies      
		/// </summary>
		/// <param name="login"> </param>
		/// <param name="password"></param>
		/// <returns>  cookies</returns>
		private CookieCollection GetBaseCookie (string login, string password)
		{
			string lj_login_chal = GetChallenge ();

			//       challenge / response
			string auth_response = GetAuthResponse (password, lj_login_chal);

			//      
			string textRequest = string.Format ("chal={0}&response={1}&user={2}",
				HttpUtility.UrlEncode (lj_login_chal),
				HttpUtility.UrlEncode (auth_response),
				HttpUtility.UrlEncode (login));
			 
			//     
			_log.WriteLine ("\r\n*** Request:");
			_log.WriteLine (textRequest);

			byte[] byteArray = Encoding.UTF8.GetBytes (textRequest);

			//   
			HttpWebRequest request =
				(HttpWebRequest)WebRequest.Create ("http://www.livejournal.com/login.bml");

			//   
			request.Method = "POST";

			//   ,   cookies,      
			request.AllowAutoRedirect = false;
			request.ContentType = "application/x-www-form-urlencoded";
			request.ContentLength = textRequest.Length;
			request.UserAgent = "LJServerTest";

			//     cookie    
			request.CookieContainer = new CookieContainer ();

			//   
			Stream requestStream = request.GetRequestStream ();
			requestStream.Write (byteArray, 0, textRequest.Length);

			//   
			HttpWebResponse response = (HttpWebResponse)request.GetResponse ();

			//  
			Stream responseStream = response.GetResponseStream ();
			StreamReader readStream = new StreamReader (responseStream, Encoding.UTF8);

			string currResponse = readStream.ReadToEnd ();

			//     
			_log.WriteLine ("\r\n*** Response:");
			_log.WriteLine (currResponse);

			readStream.Close ();
			response.Close ();

			//     cookies
			CookieCollection newCollection = new CookieCollection ();
			for (int i = 0; i < response.Cookies.Count; i++)
			{
				if (response.Cookies[i].Name == "ljloggedin" ||
					response.Cookies[i].Name == "ljmastersession")
				{
					newCollection.Add (response.Cookies[i]);
				}
			}

			//      cookie
			_log.WriteLine ("\r\n*** Cookies:");
			for (int i = 0; i < newCollection.Count; i++)
			{
				_log.WriteLine (newCollection[i].ToString ());
			}

			return newCollection;
		}

		/// <summary>
		///  md5.   lj.net. http://lj-net.cvs.sourceforge.net/lj-net/lj-net/Utils.cs?view=markup
		/// </summary>
		/// <param name="text"></param>
		/// <returns></returns>
		private string ComputeMD5 (string text)
		{
			MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider ();
			byte[] hashBytes = md5.ComputeHash( Encoding.UTF8.GetBytes( text ) );

			StringBuilder sb = new StringBuilder();

			foreach( byte hashByte in hashBytes )
				sb.Append( Convert.ToString( hashByte, 16 ).PadLeft( 2, '0' ) );

			return sb.ToString();
		}

		/// <summary>
		///    
		/// </summary>
		/// <param name="password"></param>
		/// <param name="challenge"></param>
		/// <returns></returns>
		private string GetAuthResponse (string password, string challenge)
		{
			// md5  
			string hpass = ComputeMD5 (password);

			string constr = challenge + hpass;
			string auth_response = ComputeMD5 (constr);

			return auth_response;
		}


		/// <summary>
		///  
		/// </summary>
		/// <param name="url">,    </param>
		/// <param name="textRequest"> </param>
		/// <param name="cookies">,     </param>
		/// <returns></returns>
		private string SendRequest (string url, 
			string textRequest, 
			CookieCollection cookies)
		{
			//     
			_log.WriteLine ("\r\n*** Request:");
			_log.WriteLine (textRequest);
			
			//      byte[]
			byte[] byteArray = Encoding.UTF8.GetBytes (textRequest);

			//   
			HttpWebRequest request = (HttpWebRequest)WebRequest.Create (url);

			//   
			request.Credentials = CredentialCache.DefaultCredentials;
			request.Method = "POST";
			request.ContentLength = byteArray.Length;
			request.ContentType = "application/x-www-form-urlencoded";
			request.AllowAutoRedirect = false;

			//     cookie    
			request.CookieContainer = new CookieContainer ();

			if (cookies != null)
			{
				request.CookieContainer.Add (cookies);
			}

			//   
			Stream requestStream = request.GetRequestStream ();
			requestStream.Write (byteArray, 0, byteArray.Length);

			//   
			HttpWebResponse response = (HttpWebResponse)request.GetResponse ();

			//  
			Stream responseStream = response.GetResponseStream ();
			StreamReader readStream = new StreamReader (responseStream, Encoding.UTF8);

			string currResponse = readStream.ReadToEnd ();

			//     
			_log.WriteLine ("\r\n*** Response:");
			_log.WriteLine (currResponse);

			//       cookie
			_log.WriteLine ("\r\n*** Cookies:");
			for (int i = 0; i < response.Cookies.Count; i++)
			{
				_log.WriteLine (response.Cookies[i].ToString());
			}

			readStream.Close ();
			response.Close ();

			return currResponse;
		}



		/// <summary>
		///  
		/// </summary>
		/// 
		/// <param name="login"> , 
		///      </param>
		/// 
		/// <param name="password"> </param>
		/// 
		/// <param name="url"> ,     </param>
		/// 
		/// <param name="message"> </param>
		/// 
		/// <param name="subj"> . 
		///    </param>
		/// 
		/// <param name="parent"> ,    . 
		///       , 
		///       0</param>
		/// 
		public void PostComment (string login, 
			string password, 
			string url,
			string message, 
			string subj, 
			int parent)
		{
			//     
			UrlInfo urlElements;

			try
			{
				urlElements = new UrlInfo (url);
			}
			catch (FormatException e)
			{
				_log.WriteLine (e.Message);
				throw;
			}

			_log.WriteLine ("Parsing URL...");
			_log.WriteLine (string.Format ("Username: {0}", urlElements.Journal));
			_log.WriteLine (string.Format ("Post ID: {0}", urlElements.Id));

			CookieCollection cookies = GetBaseCookie (login, password);

			//      
			string text;

			try
			{
				text = GetPage (url, cookies);
			}
			catch (WebException e)
			{
				_log.WriteLine (e.ToString());
				throw new FormatException ("URL Error", e);
			}


			string lj_form_auth = GetLjFormAuth (text);

			//      '-'  '_'
			string cookieuser = login.Replace ('-', '_');
			string journal = urlElements.Journal.Replace ('-', '_');

			//  
			StringBuilder sb = new StringBuilder ();
			sb.AppendFormat ("usertype=cookieuser&");
			sb.AppendFormat ("subject={0}&", HttpUtility.UrlEncode (subj));
			sb.AppendFormat ("body={0}&", HttpUtility.UrlEncode (message));
			sb.AppendFormat ("lj_form_auth={0}&", HttpUtility.UrlEncode (lj_form_auth));
			sb.AppendFormat ("cookieuser={0}&", cookieuser);
			sb.AppendFormat ("journal={0}&", journal);
			sb.AppendFormat ("itemid={0}&", urlElements.Id);
			sb.AppendFormat ("parenttalkid={0}&", parent);

			string request = sb.ToString ();

			//      
			string urlServer = "http://www.livejournal.com/talkpost_do.bml";

			SendRequest (urlServer, request, cookies);
		}

		private string GetLjFormAuth (string text)
		{
			_log.WriteLine ("");
			_log.WriteLine ("Parsing Form...");


			//   lj_form_auth  
			string authFormRe = 
				"(\\\\)?\"lj_form_auth(\\\\)?\" value=(\\\\)?\"(?<auth>.*?)(\\\\)?\"";

			string lj_form_auth =
				Regex.Match (text, 
				authFormRe, 
				RegexOptions.Multiline | RegexOptions.IgnoreCase).Groups["auth"].Value;


			_log.WriteLine (string.Format ("auth: {0}", lj_form_auth));

			if (lj_form_auth.Length == 0)
			{
				throw new FormatException ();
			}
			return lj_form_auth;
		}

	}	// public class LJServer
}
